package mpi.eudico.client.annotator.commands;

import mpi.eudico.client.annotator.Constants;
import mpi.eudico.client.annotator.util.AnnotationDataRecord;
import mpi.eudico.client.annotator.util.ClientLogger;

import mpi.eudico.server.corpora.clom.Annotation;
import mpi.eudico.server.corpora.clom.Tier;
import mpi.eudico.server.corpora.clom.Transcription;
import mpi.eudico.server.corpora.clom.TranscriptionStore;

import mpi.eudico.server.corpora.clomimpl.abstr.AbstractAnnotation;
import mpi.eudico.server.corpora.clomimpl.abstr.AlignableAnnotation;
import mpi.eudico.server.corpora.clomimpl.abstr.TierImpl;
import mpi.eudico.server.corpora.clomimpl.abstr.TranscriptionImpl;
import mpi.eudico.server.corpora.clomimpl.dobes.ACMTranscriptionStore;
import mpi.eudico.server.corpora.clomimpl.type.LinguisticType;

import mpi.eudico.util.TimeFormatter;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Vector;

import javax.swing.JOptionPane;


/**
 * A command that creates annotations based on the gaps between annotations on
 * a given tier. The new annotations can either be inserted on the same tier
 * or created on an other tier.
 *
 * @author Han Sloetjes
 * @author Jeffrey Lemein
 * @version August, 2010
 */
public class AnnotationsFromOverlapsCommand implements UndoableCommand, ClientLogger {
    private String commandName;
    private TranscriptionImpl transcription;
    private String[] sourceTiers;
    private String destTierName;
    private LinguisticType destLingType;
    private int annotationValueType;
    private String timeFormat;
    private String annWithValue;
    private String annFromTier;
    private int overlapsCriteria;
    private List<String[]> tierValueConstraints;
    private int transcriptionMode;
    private TranscriptionImpl[] filenames;
    private boolean usePalFormat;
    private String parentTierName;
    
    private static final int OVERLAP_CRITERIA_OVERLAP = 0;
    private static final int OVERLAP_CRITERIA_VALUES_EQUAL = 1;
    private static final int OVERLAP_CRITERIA_VALUES_NOT_EQUAL = 2;
    private static final int OVERLAP_CRITERIA_USE_SPECIFIED_CONSTRAINTS = 3;
    private boolean destTierCreated = false;
    private List<AnnotationDataRecord> createdAnnos;

    /**
     * Constructor.
     *
     * @param name the name of the command
     */
    public AnnotationsFromOverlapsCommand(String name) {
        commandName = name;
    }

    /**
     * Recreates the new tier (if applicable) and the newly created
     * annotations.
     */
    public void redo() {
        if (transcription != null) {
            TierImpl dt = (TierImpl) transcription.getTierWithId(destTierName);

            if( dt == null ){
            	dt = (TierImpl) transcription.getTierWithId(sourceTiers[0]);
                LinguisticType lType = dt.getLinguisticType();

                if (lType.getConstraints() != null) {
                    // the source tier should be a toplevel tier
                    ClientLogger.LOG.severe("The source tier is not a root tier.");

                    return;
                }

                TierImpl dTier = new TierImpl(null, destTierName,
                        dt.getParticipant(), transcription, lType);
                dTier.setAnnotator(dt.getAnnotator());
                dTier.setDefaultLocale(dt.getDefaultLocale());
                transcription.addTier(dTier);
                dt = dTier;
            }

            if (dt == null) {
                ClientLogger.LOG.severe(
                    "Could not find the destination tier for redo");

                return;
            }

            if ((createdAnnos == null) && (createdAnnos.size() == 0)) {
                ClientLogger.LOG.info("No annotations to restore");

                return;
            }

            int curPropMode = 0;

            curPropMode = transcription.getTimeChangePropagationMode();

            if (curPropMode != Transcription.NORMAL) {
                transcription.setTimeChangePropagationMode(Transcription.NORMAL);
            }

            transcription.setNotifying(false);

            AnnotationDataRecord record;
            Annotation ann;

            for (int i = 0; i < createdAnnos.size(); i++) {
                record = (AnnotationDataRecord) createdAnnos.get(i);
                ann = dt.createAnnotation(record.getBeginTime(),
                        record.getEndTime());

                if ((ann != null) && (record.getValue() != null)) {
                    ann.setValue(record.getValue());
                }
            }

            transcription.setNotifying(true);
            // restore the time propagation mode
            transcription.setTimeChangePropagationMode(curPropMode);
        }
    }

    /**
     * Deletes the new annotations and/or the new tier.
     */
    public void undo() {
        if (transcription != null) {
        	//if during the gap creation process, a new tier was added
        	if( destTierCreated ){
        		Tier dt = transcription.getTierWithId(destTierName);
        		
        		//remove that tier now
                if (dt != null) 
                    transcription.removeTier(dt);
            } else {
            	//annotations were added to an already existing tier, so remove annotations only
                TierImpl st = (TierImpl) transcription.getTierWithId(destTierName);

                if (st != null) {
                    if ((createdAnnos != null) && (createdAnnos.size() > 0)) {
                        transcription.setNotifying(false);

                        AnnotationDataRecord record;
                        Annotation ann;

                        for (int i = 0; i < createdAnnos.size(); i++) {
                            record = (AnnotationDataRecord) createdAnnos.get(i);
                            ann = st.getAnnotationAtTime((record.getBeginTime() +
                                    record.getEndTime()) / 2);

                            if (ann != null) {
                                st.removeAnnotation(ann);
                            }
                        }

                        transcription.setNotifying(true);
                    }
                }
            }
        }
    }

    /**
     * <b>Note: </b>it is assumed the types and order of the arguments are correct.<br>
     * If arg[1] is null it is assumed that the annotations have to be
     * inserted on the source tier.
     *
     * @param receiver the TranscriptionImpl
     * @param arguments the arguments: <ul>
     * 			<li>arg[0] = the selected tiers (String[])</li> 
     * 			<li>arg[1] = the new tier name (String)</li> 
     *        	<li>arg[2] = the linguistic type of the new tier (LinguisticType)</li>
     *        	<li>arg[3] = the kind of value the new tier should have (Integer)
     *        					0 = time format (then it uses arg[4])
     *        					1 = specific value (then it uses arg[5])
     *        					2 = value from another tier (then it used arg[6]) </li>
     *        	<li>arg[4] = the time format to use (String)</li> 
     *        	<li>arg[5] = the specific value to use (String)</li>
     *        	<li>arg[6] = the specific value from a tier to use (String)</li>
     *        	<li>arg[7] = the tier name of which the values should be taken (String)</li>
     *        	<li>arg[8] = the overlaps criteria (0 = same value, 1 = different value, 3 = based on constraints) (List<String[]>)
     *        	<li>arg[9] = create overlaps on currently opened transcription (0 = current transcription, 1 = based on files in arg[10]) (Integer)
     *        	<li>arg[10] = names of the files to create annotations for (String[])
     *        	<li>arg[11] = boolean value indicating if pal mode should be used (value is true) or NTSC (value is false). Only needed when time format is SMPTE.
     *        	<li>arg[12] = parent tier name (if tier is selected as parent) or null</li>
     *        </ul>
     */
    public void execute(Object receiver, Object[] arguments) {
        transcription = (TranscriptionImpl) receiver;
        
        //copy tierlist in arguments[0] into array sourceTiers
        Object[] objectArray = (Object[])arguments[0]; 
        sourceTiers = new String[objectArray.length];
        for( int i=0; i<objectArray.length; i++ )
        	sourceTiers[i] = objectArray[i].toString();
        
        destTierName = (String) arguments[1];
        destLingType = (LinguisticType) arguments[2];
        annotationValueType = (Integer) arguments[3];
        timeFormat = (String) arguments[4];
        annWithValue = (String) arguments[5];
        annFromTier = (String) arguments[6];
        overlapsCriteria = (Integer) arguments[7];
        tierValueConstraints = (List<String[]>) arguments[8];
        transcriptionMode = (Integer) arguments[9];
        usePalFormat = (Boolean) arguments[11];
        parentTierName = (String) arguments[12];
        
        Object[] filenamesObj = null;
        if( arguments[10] != null ){
	        filenamesObj = (Object[]) arguments[10];
	        filenames = new TranscriptionImpl[filenamesObj.length];
	        for( int i=0; i<filenamesObj.length; i++ )
	        	filenames[i] = (TranscriptionImpl) filenamesObj[i];
        }

        createdAnnos = new ArrayList<AnnotationDataRecord>();
        
        //create annotations
        if( transcriptionMode == 0 )
    		//use currently opened transcription
    		createAnnotations(transcription);
    	else if( transcriptionMode == 1 ){
    		//transcription based on files
    		for( TranscriptionImpl t : filenames ){
    			createAnnotations(t);
    			String directoryToSave = t.getPathName();

    			try{
    				ACMTranscriptionStore.getCurrentTranscriptionStore().storeTranscriptionIn(t, null, new Vector(), directoryToSave, TranscriptionStore.EAF);
    			}catch(IOException e){
					LOG.warning("Can not write transcription to file with directory/filename: " + directoryToSave );
    			}
    		}
    	}        
    }

    /**
     * Returns the name of the command.
     *
     * @return the name of the command
     */
    public String getName() {
        return commandName;
    }

    /**
     * Iterates over the annotations of the source tier and creates annotations
     * for the gaps.
     */
    private void createAnnotations(TranscriptionImpl t) {
    	//retrieve original tier (the tiers with gaps)    	
    	ArrayList<TierImpl> tierList = new ArrayList<TierImpl>();    		
    	TierImpl parentTier = (TierImpl)t.getTierWithId(parentTierName);
    	
    	for( int i=0; i < sourceTiers.length; i++ ){
    		TierImpl tier = (TierImpl) t.getTierWithId( sourceTiers[i] );
    		
    		//check if source tier is present
    		if (tier == null)
                return;
    		
    		tierList.add(tier);
    	}

    	//check if the selected tier(it's demanded)
        if (parentTier == null && destLingType.getConstraints() != null) {
        	// the source tier should be a toplevel tier
        	ClientLogger.LOG.severe("Tier to be created has an invalid linguistic type (tier should not be a root tier)");

            return;
        }
        
        TierImpl dTier = new TierImpl(parentTier, destTierName, tierList.get(0).getParticipant(), t, destLingType);
        dTier.setAnnotator(tierList.get(0).getAnnotator());
        dTier.setDefaultLocale(tierList.get(0).getDefaultLocale());
        t.addTier(dTier);

        //flag for the undo command that the destination tier is created so that
        //when undo takes place, the command knows that we can remove the tier as a whole
        destTierCreated = true;

        //Stage 2: Computing overlaps
        //         dTier contains name of tier where the gaps should be placed
        //         tierList contains the tiers where the gaps should be computed from
        int curPropMode = t.getTimeChangePropagationMode();
        t.setNotifying(false);

        if (curPropMode != Transcription.NORMAL) {
            t.setTimeChangePropagationMode(Transcription.NORMAL);
        }

        //get the annotations from the source tier
        ArrayList<Overlap> commonOverlaps = getCommonOverlaps( tierList );
        AlignableAnnotation aa = null;
        int numAnnCreated = 0;
        
        for( int i=0; i<commonOverlaps.size(); i++ ){
        	Overlap overlap = commonOverlaps.get(i);
        	if (overlap.end <= overlap.begin) {
        		ClientLogger.LOG.warning("Cannot create annotation, begin time >= end time: " + overlap.begin + " - " + overlap.end);
        		continue;
        	}
        	aa = (AlignableAnnotation) dTier.createAnnotation(overlap.begin, overlap.end);
        	
        	if (aa != null) {
        		switch( annotationValueType ){
	        		//time format
        			case 0:
        				aa.setValue(getTimeString(overlap.end-overlap.begin));
	        			break;
	        		
	        		//specific value
	        		case 1:
	        			aa.setValue(annWithValue);
	        			break;
	        		
	        		//value from tier
	        		case 2:
	        			TierImpl tier = (TierImpl) t.getTierWithId( annFromTier );
	        			aa.setValue(tier.getAnnotationAfter((overlap.begin + overlap.end) / 2).getValue());
	        			break;
	        			
	        		// concat the values
	        		case 3:
	        			aa.setValue(overlap.value);
//	        			break;
        		}
        		numAnnCreated++;
                createdAnnos.add(new AnnotationDataRecord(aa));
            } else {
                ClientLogger.LOG.warning("Annotation could not be created " +
                    overlap.begin + "-" + overlap.end);
            }
        }

        ClientLogger.LOG.info("Number of annotations created: " + numAnnCreated );

        if (t.getTimeChangePropagationMode() != curPropMode) {
            t.setTimeChangePropagationMode(curPropMode);
        }

        t.setNotifying(true);
    }

    /**
     * Converts a numeric time value to a (formatted) string.
     *
     * @param time
     *
     * @return a string representation
     */
    private String getTimeString(long time) {
        if (Constants.HHMMSSMS_STRING.equals(timeFormat)) {
            return TimeFormatter.toString(time);
        } else if (Constants.SSMS_STRING.equals(timeFormat)) {
            return TimeFormatter.toSSMSString(time);
        } else if (Constants.HHMMSSFF_STRING.equals(timeFormat)) {
        	if( usePalFormat )
        		return TimeFormatter.toTimecodePAL(time);
        	else
        		return TimeFormatter.toTimecodeNTSC(time);
        } else {
            return String.valueOf(time);
        }
    }
    
    /**
     * Given a list of tiers, it computes the common gaps that all the tiers in the list have.
     * A common gap is the space between two annotations (from the same or a different tier) 
     * in which no annotation occurs.
     * @param tierList the list of tiers where the common gaps are computed for
     * @return a list of gaps
     */
    private ArrayList<Overlap> getCommonOverlaps( ArrayList<TierImpl> tierList ){
    	ArrayList<Overlap> overlapList = new ArrayList<Overlap>();
    	Vector<Annotation> annotations;
    	String constraint;
    	TierImpl currentTier;
    	long timeSoFar = 0, beginTime;
    	
    	//if there are no tiers to be inspected, then there are no overlaps as well
    	if( tierList.size() <= 0 ){
    		//JOptionPane.showMessageDialog(null, "tierList.size() <= 0");
    		return overlapList;
    	}
    	
    	/**
    	 * Stage 1: Get overlaps from first tier
    	 */
    	currentTier = tierList.get(0);
    	annotations = currentTier.getAnnotations();
    	
    	//JOptionPane.showMessageDialog(null, "Tier with tier name: " + currentTier.getName() + " has " + annotations.size() + " annotations");
    	
    	//if there are no annotations on a tier, then there can be no overlaps
    	if( annotations.size() <= 0 ){
    		//JOptionPane.showMessageDialog(null, "annotations.size() <= 0");
    		return overlapList;
    	}
    	
    	constraint = getConstraint(currentTier.getName());
    	
    	//loop over all annotations and store the overlaps
    	for( int i=0; i<annotations.size(); i++ ){
    		AbstractAnnotation abstrAnnotation = (AbstractAnnotation) annotations.get(i);
    		beginTime = abstrAnnotation.getBeginTimeBoundary();
    		long endTime = abstrAnnotation.getEndTimeBoundary();
    		
    		//begin by viewing annotations as overlaps when the overlaps criteria is different than 
    		//obeying user constraints. If they should obey user constraints, then add them when they
    		//are equal to the constraint or when there is no constraint specified for the tier name.
    		//JOptionPane.showMessageDialog(null, "annotation value = " + abstrAnnotation.getValue() + ", constraint = " + constraint + "\nOVERLAP_CRITERIA_USE_SPECIFIED_CONSTRAINTS = " + OVERLAP_CRITERIA_USE_SPECIFIED_CONSTRAINTS + "\nOverlapsCriteria = " + overlapsCriteria);
    		if( overlapsCriteria != OVERLAP_CRITERIA_USE_SPECIFIED_CONSTRAINTS || (constraint != null && constraint.equals(abstrAnnotation.getValue())) ){
    			overlapList.add(new Overlap(beginTime, endTime, abstrAnnotation.getValue()));
    		}
    	}
    	
    	/**
    	 * Stage 2: Loop over the remaining tiers and remove overlaps
    	 */
    	for( int i=1; i<tierList.size(); i++ ){
    		//if there are no overlaps, then no overlaps will be created
	    	if( overlapList.size() <= 0 ){
	    		//JOptionPane.showMessageDialog(null, "stage 2: overlapList.size() < 0");
	    		break;
	    	}
	    	
    		currentTier = tierList.get(i);
	    	annotations = currentTier.getAnnotations();
	    	
	    	//JOptionPane.showMessageDialog(null, "Tier with tier name: " + currentTier.getName() + " has " + annotations.size() + " annotations");
	    		
	    	constraint = getConstraint(currentTier.getName());
	    	
	    	for( int overlapNr = 0; overlapNr < overlapList.size(); overlapNr++ ){
	    		if( annotations.size() <= 0 ){
	    			for( int j=overlapList.size()-1; j>=overlapNr; j-- )
	    				overlapList.remove(j);
	    			//JOptionPane.showMessageDialog(null, "stage 2: break because annotations.size() < 0");
	    			break;
	    		}
	    		
	    		AbstractAnnotation abstrAnnotation = (AbstractAnnotation) annotations.get(0);
	    		beginTime = abstrAnnotation.getBeginTimeBoundary();
	    		long endTime = abstrAnnotation.getEndTimeBoundary();
	    		
	    		//decide to go on to compute overlaps for this annotation
	    		constraint = getConstraint(currentTier.getName());
	    		
	    		//System.out.print("checkConstraint");
	    		boolean matchConstraint = checkConstraint(overlapList.get(overlapNr).value, abstrAnnotation.getValue(), overlapsCriteria, constraint);
	    		//System.out.println(" - finished (matchConstraint = " + matchConstraint + ")");
	    		
	    		//String value1 = overlapList.get(overlapNr).value;
	    		//String value2 = abstrAnnotation.getValue();
	    		
	    		if( !matchConstraint ){
	    			//JOptionPane.showMessageDialog(null, value1 + " and " + value2 + " does not match constraint");
	    			
	    			if( endTime < overlapList.get(overlapNr).end ){
	    				annotations.remove(0);	    				
	    			}else{
	    				overlapList.remove(overlapNr);
	    			}
	    			overlapNr--;
	    			continue;
	    		}
	    		
	    		//System.out.println("!matchConstraint");
	    		
	    		//if annotation comes before the calculated overlaps
	    		//      |---|
	    		//|---|
	    		if( endTime < overlapList.get(overlapNr).begin ){
	    			//JOptionPane.showMessageDialog(null, "annotation " + value2 + " comes before first annotation " + value1 + ", so remove annotation");
	    			annotations.remove(0);	    			
	    			overlapNr--;
	    			continue;
	    		}
	    		
	    		//  |---|
	    		//|---|
	    		if( beginTime < overlapList.get(overlapNr).begin && endTime <= overlapList.get(overlapNr).end ){
	    			//JOptionPane.showMessageDialog(null, "annotation " + value2 + " starts before and ends during first annotation " + value1 + ", update overlap and remove annotation");
	    			if( endTime < overlapList.get(overlapNr).end ){
	    				overlapList.add(overlapNr, new Overlap(overlapList.get(overlapNr).begin, endTime, abstrAnnotation.getValue()+" "+overlapList.get(overlapNr).value ));
	    				overlapList.get(overlapNr+1).begin = endTime;
	    			}else{
	    				overlapList.get(overlapNr).end = endTime;
	    				overlapList.get(overlapNr).value = abstrAnnotation.getValue()+" " + overlapList.get(overlapNr).value;
	    			}
	    			annotations.remove(0);	    			
	    			continue;
	    		}
	    		
	    		//|-------|
	    		//  |---|
	    		if( beginTime >= overlapList.get(overlapNr).begin && endTime <= overlapList.get(overlapNr).end ){
	    			//JOptionPane.showMessageDialog(null, "annotation " + value2 + " starts and ends during first annotation " + value1 + ", update overlap and remove annotation");
	    			if( endTime < overlapList.get(overlapNr).end ){
	    				overlapList.add(overlapNr, new Overlap(beginTime, endTime, overlapList.get(overlapNr).value +" "+ abstrAnnotation.getValue()));
	    				overlapList.get(overlapNr+1).begin = endTime;
	    			}else{
	    				overlapList.get(overlapNr).begin = beginTime;
	    				overlapList.get(overlapNr).value = overlapList.get(overlapNr).value +" "+ abstrAnnotation.getValue();
	    			}
	    			annotations.remove(0);
	    			continue;
	    		}
	    		
	    		//|----|
	    		//  |------|
	    		if( beginTime >= overlapList.get(overlapNr).begin && endTime > overlapList.get(overlapNr).end ){
	    			//JOptionPane.showMessageDialog(null, "annotation " + value2 + " starts during and ends after first annotation " + value1 + ", update overlap");
	    			overlapList.get(overlapNr).begin = beginTime;
	    			overlapList.get(overlapNr).value = overlapList.get(overlapNr).value +" "+ abstrAnnotation.getValue();
	    			continue;
	    		}
	    		
	    		//   |----|
	    		//|----------|
	    		if( beginTime < overlapList.get(overlapNr).begin && endTime > overlapList.get(overlapNr).end ){
	    			//JOptionPane.showMessageDialog(null, "annotation " + value2 + " starts before and ends after first annotation " + value1 + ", so do nothing");
	    			overlapList.get(overlapNr).value = abstrAnnotation.getValue() +" "+ overlapList.get(overlapNr).value ;
	    			continue;
	    		}
	    		
	    		//|----|
	    		//      |--|
	    		if( beginTime >= overlapList.get(overlapNr).end ){
	    			//JOptionPane.showMessageDialog(null, "annotation " + value2 + " starts after first annotation " + value1 + " , remove overlap (ERROR probably)");
	    			overlapList.remove(overlapNr);
	    			overlapNr--;
	    			continue;
	    		}
	    	}
    	}
    	
    	return overlapList;
    }
    
    /**
     * Checks if the overlap value match the specified constraint
     * @param overlapValue the value of the overlap to be checked
     * @param newValue the value of the annotation for which the constraint check is performed
     * @param overlapsCriteria the criteria used to check if constraints are obeyed (overlap, equal value, inequal value, or user specified constraints.
     * @param constraint the value that newValue should have to obey the constraint.
     * @return
     */
    private boolean checkConstraint(String overlapValue, String newValue, int overlapsCriteria, String constraint){
    	//the only constraint is that it should be an overlap, so return true because values are not relevant
    	if( overlapsCriteria == OVERLAP_CRITERIA_OVERLAP )
    		return true;
    	
    	if( overlapsCriteria == OVERLAP_CRITERIA_VALUES_EQUAL )
    		return overlapValue.equals(newValue);
    	
    	if( overlapsCriteria == OVERLAP_CRITERIA_VALUES_NOT_EQUAL )
    		return !overlapValue.equals(newValue);    	
    	
    	//use the specified constraints by the user
    	if( overlapsCriteria == OVERLAP_CRITERIA_USE_SPECIFIED_CONSTRAINTS )
    		return constraint == null || newValue.equals(constraint);
    	
    	return true;
    }
    
    /**
     * Returns the specified constraint for the given tier name.
     * @param tierName the name of the tier for which the constraint should be given.
     * @return the value (constraint) that the given tier name should have.
     * or NULL if no constraints are specified for this tier name.
     */
    private String getConstraint(String tierName) {
    	//System.out.println("Checking if tier " + tierName + " has a constraint");
    	for( int i=0; i<tierValueConstraints.size(); i++ ){
    		//System.out.println("tiervalueConstraints[0] = " + tierValueConstraints.get(i)[0]);
    		
			if( tierValueConstraints.get(i)[0].equals(tierName) ){
				//System.out.println("READY: return value");
				return tierValueConstraints.get(i)[1];
			}
    	}
    	
		//System.out.println("READY: return null");
		return null;
	}
    
    /**
     * Class that represents an overlap
     * @author Jeffrey Lemein
     * @version March, 2011
     */
    private class Overlap{
    	public long begin, end;
    	public String value;
    	
    	public Overlap( long begin, long end ){
    		this.begin = begin;
    		this.end = end;
    	}
    	
    	public Overlap( long begin, long end, String value ){
    		this(begin, end);
    		this.value = value;
    	}
    }
}
