package mpi.eudico.client.annotator.commands;

import java.io.File;
import java.util.List;

import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.filechooser.FileFilter;

import mpi.eudico.client.annotator.ElanLocale;
import mpi.eudico.client.annotator.Preferences;
import mpi.eudico.client.annotator.recognizer.data.Segmentation;
import mpi.eudico.client.annotator.recognizer.io.CsvTierIO;
import mpi.eudico.client.annotator.recognizer.io.XmlTierIO;
import mpi.eudico.client.annotator.util.ElanFileFilter;
import mpi.eudico.client.annotator.util.FileExtension;
import mpi.eudico.server.corpora.clomimpl.abstr.TranscriptionImpl;

/**
 * A Command to load tiers from a recognizer tier file. 
 * @author Han Sloetjes
 *
 */
public class ImportRecogTiersCommand implements Command {
	private String name;
	private TranscriptionImpl trans;
	
	/**
	 * Constructor.
	 * @param name name of the command
	 */
	public ImportRecogTiersCommand(String name) {
		this.name = name;
	}

	/**
	 * @param receiver the transcription
	 * @param arguments null
	 */
	public void execute(Object receiver, Object[] arguments) {
		trans = (TranscriptionImpl) receiver;

		String filePath = promptForTierFile();
		
		File f = new File(filePath);
        if (f.exists() && f.canRead()) {
        	List<Segmentation> segm = null;
        	
        	if (filePath.endsWith("csv")) {
				CsvTierIO cio = new CsvTierIO();
				segm = cio.read(f);
        	} else {
				XmlTierIO xio = new XmlTierIO(f);
				segm = xio.parse();
        	}
        	
        	if (segm != null && segm.size() > 0) {
                Command cc = ELANCommandFactory.createCommand(trans,
                        ELANCommandFactory.SEGMENTS_2_TIER_DLG);
                cc.execute(trans,
                    new Object[] { segm });
        	} else {
        		// warning no tiers 
            	JOptionPane.showMessageDialog(ELANCommandFactory.getRootFrame(trans), 
            			ElanLocale.getString("Recognizer.RecognizerPanel.Warning.NoTiers"), 
            			ElanLocale.getString("Message.Error"), JOptionPane.WARNING_MESSAGE);
        	}
        } else {
        	// warning file read failed
        	JOptionPane.showMessageDialog(ELANCommandFactory.getRootFrame(trans), 
        			ElanLocale.getString("Recognizer.RecognizerPanel.Warning.LoadFailed"), 
        			ElanLocale.getString("Message.Error"), JOptionPane.WARNING_MESSAGE);
        }
	}

	public String getName() {
		return name;
	}

	/**
	 * Prompts the user to specify a location from where to load tiers.
	 * The tiers can be in a csv file or in an xml file
	 * 
	 * @return the path or null if canceled
	 */
	private String promptForTierFile() {
		String prefPath = null;
		
		Object val = Preferences.get("Recognizer.Path", null);
		
		if (val instanceof String) {
			prefPath = (String) val;
		}
		
		JFileChooser chooser = null;
		if (prefPath != null) {
			chooser = new JFileChooser(prefPath);
		} else {
			chooser = new JFileChooser();
		}
		FileFilter ff =  ElanFileFilter.createFileFilter(ElanFileFilter.CSV_TYPE);
		FileFilter ff2 = ElanFileFilter.createFileFilter(ElanFileFilter.XML_TYPE);
		
		
		chooser.addChoosableFileFilter(ff);
		chooser.addChoosableFileFilter(ff2);
		chooser.setFileFilter(ff2);
		
		if (chooser.showOpenDialog(ELANCommandFactory.getRootFrame(trans)) == JFileChooser.APPROVE_OPTION) {
			File f = chooser.getSelectedFile();
			File par = f.getParentFile();
			if (par!= null && par.isDirectory()) {
				Preferences.set("Recognizer.Path", par.getAbsolutePath(), null);
			}
			
			if (chooser.getFileFilter() == ff) {
				String lower = f.getAbsolutePath().toLowerCase();
				boolean valid = false;
				for (int i = 0; i < FileExtension.CSV_EXT.length; i++) {
					if (lower.endsWith(FileExtension.CSV_EXT[i])) {
						valid = true;
						break;
					}
				}
				if (!valid) {
					f = new File(f.getAbsolutePath() + "." + FileExtension.CSV_EXT[0]);
				}
			} else {
				if (!f.getAbsolutePath().toLowerCase().endsWith("xml")) {
					f = new File(f.getAbsolutePath() + ".xml");
				}
			}
			return f.getAbsolutePath();
		} else {
			return null;
		}		
	}
}
