package mpi.eudico.client.annotator.export;

import java.awt.Frame;
import java.awt.GridBagConstraints;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Vector;

import javax.swing.DefaultCellEditor;
import javax.swing.JCheckBox;
import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.border.TitledBorder;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.filechooser.FileFilter;

import mpi.eudico.client.annotator.ElanLocale;
import mpi.eudico.client.annotator.Preferences;
import mpi.eudico.client.annotator.Selection;
import mpi.eudico.client.annotator.commands.ELANCommandFactory;
import mpi.eudico.client.annotator.recognizer.data.AudioSegment;
import mpi.eudico.client.annotator.recognizer.data.RSelection;
import mpi.eudico.client.annotator.recognizer.data.Segment;
import mpi.eudico.client.annotator.recognizer.data.Segmentation;
import mpi.eudico.client.annotator.recognizer.gui.SelectionPanel;
import mpi.eudico.client.annotator.recognizer.io.RecTierWriter;
import mpi.eudico.client.annotator.util.ElanFileFilter;
import mpi.eudico.client.annotator.util.FileExtension;
import mpi.eudico.client.util.CheckBoxTableCellRenderer;
import mpi.eudico.server.corpora.clom.Transcription;
import mpi.eudico.server.corpora.clomimpl.abstr.AbstractAnnotation;
import mpi.eudico.server.corpora.clomimpl.abstr.TierImpl;
import mpi.eudico.server.corpora.clomimpl.praat.PraatTGEncoderInfo;
import mpi.eudico.server.corpora.clomimpl.praat.PraatTextGridEncoder;
import mpi.eudico.util.TimeRelation;
import mpi.eudico.client.annotator.gui.FileAndEncodingChooser;

/**
 * A dialog to export a selection of tiers to the AVATecH TIER format.
 * 
 * @author Han Sloetjes
 */
public class ExportRecogTiersDialog extends AbstractExtTierExportDialog 
    implements ActionListener, ItemListener, ListSelectionListener{
    private JCheckBox rootTiersCB;
    private JCheckBox selectionCB;
    
    /**
     * @param parent the parent frame
     * @param modal the modal flag
     * @param transcription the transcription
     */
    public ExportRecogTiersDialog(Frame parent, boolean modal,
            Transcription transcription, Selection selection) {
        super(parent, modal, transcription, selection);

        makeLayout();
        extractTiers();
        postInit();
    }

    /**
     * The action performed event handling.
     *
     * @param ae the action event
     */
    public void actionPerformed(ActionEvent ae) {
        super.actionPerformed(ae);
    }

    /**
     * The item state changed handling.
     *
     * @param ie the ItemEvent
     */
    public void itemStateChanged(ItemEvent ie) {
        extractTiers();
    }

    /**
     * Updates the checked state of the export checkboxes.
     *
     * @param lse the list selection event
     */
    public void valueChanged(ListSelectionEvent lse) {
        if ((model != null) && lse.getValueIsAdjusting()) {
            int b = lse.getFirstIndex();
            int e = lse.getLastIndex();
            int col = model.findColumn(EXPORT_COLUMN);

            for (int i = b; i <= e; i++) {
                if (tierTable.isRowSelected(i)) {
                    model.setValueAt(Boolean.TRUE, i, col);
                }
            }
        }
    }

    /**
     * Extract candidate tiers for export.
     */
    protected void extractTiers() {
        if (model != null) {
            for (int i = model.getRowCount() - 1; i >= 0; i--) {
                model.removeRow(i);
            }

            if (transcription != null) {
                Vector v = transcription.getTiers();
                TierImpl t;
                boolean rootsOnly = rootTiersCB.isSelected();

                for (int i = 0; i < v.size(); i++) {
                    t = (TierImpl) v.get(i);

                    if (rootsOnly) {
                        // add only root tiers
                        if (t.getParentTier() == null) {
                            model.addRow(new Object[] { Boolean.TRUE, t.getName() });
                        }
                    } else {
                        // add all
                        if (t.getParentTier() == null) {
                            model.addRow(new Object[] { Boolean.TRUE, t.getName() });
                        } else {
                            model.addRow(new Object[] { Boolean.FALSE, t.getName() });
                        }
                    }
                }
            }
            
            //read Preferences
            Object useTyp = Preferences.get("ExportRecogTiersDialog.selectedTiers", transcription);
            if (useTyp instanceof ArrayList) {
            	loadTierPreferences((ArrayList)useTyp);        	
          	 } 

            if (model.getRowCount() > 1) {
                upButton.setEnabled(true);
                downButton.setEnabled(true);
            } else {
                upButton.setEnabled(false);
                downButton.setEnabled(false);
            }
        } else {
            upButton.setEnabled(false);
            downButton.setEnabled(false);
        }
    }

    
    /**
     * Initializes UI elements.
     */
    protected void makeLayout() {
        super.makeLayout();
        rootTiersCB = new JCheckBox();
        selectionCB = new JCheckBox();
        
        GridBagConstraints gridBagConstraints;

        rootTiersCB.setSelected(true);
        rootTiersCB.addItemListener(this);
        gridBagConstraints = new GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 0;
        gridBagConstraints.gridwidth = 2;
        gridBagConstraints.anchor = GridBagConstraints.WEST;
        gridBagConstraints.insets = insets;
        tierSelectionPanel.add(rootTiersCB, gridBagConstraints);

        model.setColumnIdentifiers(new String[] { EXPORT_COLUMN, TIER_NAME_COLUMN });

        tierTable.getColumn(EXPORT_COLUMN).setCellEditor(new DefaultCellEditor(
                new JCheckBox()));
        tierTable.getColumn(EXPORT_COLUMN).setCellRenderer(new CheckBoxTableCellRenderer());
        tierTable.getColumn(EXPORT_COLUMN).setMaxWidth(30);
        tierTable.setShowVerticalLines(false);
        tierTable.setTableHeader(null);
        tierTable.getSelectionModel().addListSelectionListener(this);
    
        gridBagConstraints = new GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 0;
        gridBagConstraints.anchor = GridBagConstraints.WEST;
        gridBagConstraints.fill = GridBagConstraints.HORIZONTAL;
        gridBagConstraints.weightx = 1.0;
        gridBagConstraints.insets = insets;
        optionsPanel.add(selectionCB, gridBagConstraints);

        setPreferredSetting();
        updateLocale();
    }
    
    /**
     * Starts the actual export after performing some checks.
     *
     * @return true if export succeeded, false otherwise
     * @see mpi.eudico.client.annotator.export.AbstractBasicExportDialog#startExport()
     */
    protected boolean startExport() throws IOException {
        List selectedTiers = getSelectedTiers();
        savePreferences();

        if (selectedTiers.size() == 0) {
            JOptionPane.showMessageDialog(this,
                ElanLocale.getString("ExportTradTranscript.Message.NoTiers"),
                ElanLocale.getString("Message.Warning"),
                JOptionPane.WARNING_MESSAGE);

            return false;
        }

        String fileName = promptForFileName();
        if (fileName == null) {
        	return false;
        }
		File tf = new File(fileName);
		try { 
			if (tf.exists()) {
                int answer = JOptionPane.showConfirmDialog(this,
                        ElanLocale.getString("Message.Overwrite"),
                        ElanLocale.getString("SaveDialog.Message.Title"),
                        JOptionPane.YES_NO_OPTION,
                        JOptionPane.WARNING_MESSAGE);

                if (answer == JOptionPane.NO_OPTION) {
                    return false;
                }
			}
		} catch (Exception ex) {// any exception
			return false;
		}
		
        long begin = 0l;
        long end = Long.MAX_VALUE;
        if (selectionCB.isSelected()) {
            if (selection != null && selection.getBeginTime() < selection.getEndTime()) {
                begin = selection.getBeginTime();
                end = selection.getEndTime();
            }
        }
        
        List<Object> segmentations = new ArrayList<Object>(selectedTiers.size());
		TierImpl ti;
		String name;
		List anns;
		
		for (int i = 0; i < selectedTiers.size(); i++) {
			name = (String) selectedTiers.get(i);
			ti = (TierImpl) transcription.getTierWithId(name);
			
			if (ti != null) {					
				anns = ti.getAnnotations();
				ArrayList<RSelection> segments = new ArrayList<RSelection>(anns.size());
				AbstractAnnotation aa;
				for (int j = 0; j < anns.size(); j++) {
					aa = (AbstractAnnotation) anns.get(j);
					if (TimeRelation.overlaps(aa, begin, end)) {
						segments.add(new Segment(aa.getBeginTimeBoundary(), aa.getEndTimeBoundary(), 
							aa.getValue()));
					}
					if (aa.getBeginTimeBoundary() > end) {
						break;
					}
				}
				Segmentation segmentation = new Segmentation(name, segments, "");// pass the master media path?
				segmentations.add(segmentation);
			}
		}
        
		try {
			RecTierWriter xTierWriter = new RecTierWriter();
			xTierWriter.write(tf, segmentations, false);
		} catch (IOException ioe) {
			// show message
			JOptionPane.showMessageDialog(this, ElanLocale.getString(
					"Recognizer.RecognizerPanel.Warning.SaveFailed")  + ioe.getMessage(), 
					ElanLocale.getString("Message.Warning"), JOptionPane.ERROR_MESSAGE);
			return false;
		}
		
        return true;
    }
    
    /**
     * Applies localized strings to the ui elements.
     */
    protected void updateLocale() {
        super.updateLocale();
        setTitle(ElanLocale.getString("ExportDialog.RecogTiers.Title"));
        titleLabel.setText(ElanLocale.getString("ExportDialog.RecogTiers.Title"));
        rootTiersCB.setText(ElanLocale.getString(
                "ExportTradTranscript.Label.RootTiers"));
        selectionCB.setText(ElanLocale.getString("ExportDialog.Restrict"));
    }
    
    protected String promptForFileName() {
    	String prefPath = null;

		Object val = Preferences.get("Recognizer.Path", null);
		
		if (val instanceof String) {
			prefPath = (String) val;
		} else {
			prefPath = System.getProperty("user.home");
		}

		JFileChooser chooser = null;
		if (prefPath != null) {
			chooser = new JFileChooser(prefPath);
		} else {
			chooser = new JFileChooser();
		}
		FileFilter ff = ElanFileFilter.createFileFilter(ElanFileFilter.CSV_TYPE);
		FileFilter ff2 = ElanFileFilter.createFileFilter(ElanFileFilter.XML_TYPE);
		
		chooser.addChoosableFileFilter(ff);
		chooser.addChoosableFileFilter(ff2);
		chooser.setFileFilter(ff2);
		
		if (chooser.showSaveDialog(this) == JFileChooser.APPROVE_OPTION) {
			File f = chooser.getSelectedFile();
			File par = f.getParentFile();
			if (par!= null && par.isDirectory()) {
				Preferences.set("Recognizer.Path", par.getAbsolutePath(), null);
			}
			
			if (chooser.getFileFilter() == ff) {
				String lower = f.getAbsolutePath().toLowerCase();
				boolean valid = false;
				for (int i = 0; i < FileExtension.CSV_EXT.length; i++) {
					if (lower.endsWith(FileExtension.CSV_EXT[i])) {
						valid = true;
						break;
					}
				}
				if (!valid) {
					f = new File(f.getAbsolutePath() + "." + FileExtension.CSV_EXT[0]);
				}
			} else {
				if (!f.getAbsolutePath().toLowerCase().endsWith("xml")) {
					f = new File(f.getAbsolutePath() + ".xml");
				}
			}
			return f.getAbsolutePath();
		} else {
			return null;
		}	
    }
    
    /**
     * Intializes the dialogBox with the last preferred/ used settings 
     *
     */
    private void setPreferredSetting()
    {
    	Object useTyp = Preferences.get("ExportRecogTiersDialog.rootTiersCB", null);
    
    	if(useTyp != null){
    		rootTiersCB.setSelected((Boolean)useTyp); 
    	}	
     
    	useTyp = Preferences.get("ExportRecogTiersDialog.selectionCB", null);
    	if(useTyp != null){
    		selectionCB.setSelected((Boolean)useTyp); 
    	}     
    	
    }
    
    /**
     * Saves the preferred settings Used. 
     *
     */
    private void savePreferences(){
    	Preferences.set("ExportRecogTiersDialog.rootTiersCB", rootTiersCB.isSelected(), null);    
    	Preferences.set("ExportRecogTiersDialog.selectionCB", selectionCB.isSelected(), null);
    	Preferences.set("ExportRecogTiersDialog.selectedTiers", getSelectedTiers(), transcription);
    	
    }

    
}
